# Stubs for cachecontrol (Python 2)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

from .wrapper import CacheControl as CacheControl

__email__ = ...  # type: Any
