class BaseCache:
    def get(self, key: str) -> bytes | None: ...
    def set(self, key: str, value: bytes, expires: int | None = None) -> None: ...
    def delete(self, key: str) -> None: ...
    def close(self) -> None: ...
