namespace ${project_name};

internal static class Vocabs
{

    public static readonly Dictionary<string, string> Vocab = new()
    {
        ${vocab}
    };

    public static readonly Dictionary<string, string> Rvocab = new()
    {
        ${rvocab}
    };
}
